import argparse

import os
import sys
import lxml.etree
import getpass
import datetime 
import shutil

from info import __version__, __description__

SILECS_VERSION = "2.3.0"
DESIGN_SHEMA_PATH = "../silecs-model/src/xml/DesignSchema.xsd"
DEPLOY_SHEMA_PATH = "../silecs-model/src/xml/DeploySchema.xsd"
SILECSDESIGN = "silecsdesign"
SILECSDEPLOY = "silecsdeploy"
DESIGN = "design"
DEPLOY = "deploy"
USER = getpass.getuser()

# FESA specific variables
FESA_VERSION="7.4.0"
FESA_BASE=f"/opt/fesa/fesa-fwk/{FESA_VERSION}"
FESA_XSD=f"{FESA_BASE}/fesa-model-gsi/xml/deployment/deployment-gsi.xsd"
FESA_GSI_TEMPLATE=f"{FESA_BASE}/fesa-model-gsi/xml/design/templates/GSIClassTemplate.xml"


# silecs specific variables
try:
    SILECS_BASE=(os.environ['SILECS_BASE'])
except:
    SILECS_BASE=f"/common/usr/cscofe/silecs/${SILECS_VERSION}"


SILECS_CODEGEN_BASE=f"{SILECS_BASE}/silecs-codegen/xml/"
SILECS_CODEGEN_MIGRATION=f"{SILECS_BASE}/silecs-codegen/xml/migration"
# TODO: Check for concrete fesa version -  See here for details: https://gitlab.com/al.schwinn/silecs-cli/-/issues/4
SILECS_CODEGEN_FESA=f"{SILECS_BASE}/silecs-codegen/xml/fesa/fesa_7_3_0/"


sys.path.append(SILECS_CODEGEN_BASE)
sys.path.append(SILECS_CODEGEN_MIGRATION)
sys.path.append(SILECS_CODEGEN_FESA)
import generateFesaDesign


print(f"using sileces base path: {SILECS_BASE}")


"""
    ####### Helper functions #######
"""

def get_extension(filepath):
    return filepath.split('.')[-1]


def get_project_name(filepath):
    return os.path.splitext(os.path.basename(filepath))[0]


def get_project_dir(filepath):
    return os.path.dirname(filepath)


def get_schema_path(silecs_design_deploy_path):
    extension = get_extension(silecs_design_deploy_path)
    if extension == SILECSDESIGN:
        return DESIGN_SHEMA_PATH
    elif extension == SILECSDEPLOY:
        return DEPLOY_SHEMA_PATH
    else:
        raise Exception("Error: Passed file for validation needs to have the extension '.silecsdesign' or '.silecsdeploy")


def get_silecs_new_file_path(fesa_design_deploy_path): 
    extension = get_extension(fesa_design_deploy_path)
    project_name = get_project_name(fesa_design_deploy_path)
    project_dir = get_project_dir(fesa_design_deploy_path)
    new_file = project_name 

    if extension == DESIGN:
        extension =  SILECSDESIGN
    elif extension == DEPLOY:
        extension = SILECSDEPLOY
    else: 
        raise Exception("Error: Passed FESA file needs to have the extension '.design' or '.deploy'")

    return os.path.join(project_dir, project_name + "." + extension)


def create_backup_file(filepath):
    backup_file = filepath + ".backup"
    with open(backup_file, 'w+'):
        shutil.copyfile(src=filepath, dst=backup_file)

"""
    ####### Validation #######
"""

def validate(silecs_design_deploy_path, xsd_path):
    try:
        print("validating {}".format(silecs_design_deploy_path))
        xmlschema_doc = lxml.etree.parse(xsd_path)
        xmlschema = lxml.etree.XMLSchema(xmlschema_doc)
        xml_doc = lxml.etree.parse(silecs_design_deploy_path)
        result = xmlschema.validate(xml_doc)
        return result
    except OSError as e:
        print(e)
        return False
    except Exception as e:
        print(e)
        return False

# TODO: Implement jave-rules -  See here for details: https://gitlab.com/al.schwinn/silecs-cli/-/issues/3


def silecs_validate(file_paths):
    for path in file_paths:
        try:
            xsd_path = get_schema_path(path)
            val_result = validate(path, xsd_path)
            if val_result:
                print("File is valid")
            else:
                print("File is not valid. Check errors above.")
        except Exception as e:  
            print(e)

"""
    ####### Create #######
"""

def create_silecs_design(filepath, project_name):
    date = datetime.datetime.today().strftime('%m/%d/%Y')
    design_template = f"""DESIGN_TEMPLATE='<?xml version="1.0" encoding="UTF-8"?>
<SILECS-Design silecs-version="{SILECS_VERSION}" created="{date}" updated="{date}"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:noNamespaceSchemaLocation="{DESIGN_SHEMA_PATH}">
    <Information>
        <Owner user-login="{USER}"/>
        <Editor user-login="{USER}"/>
    </Information>
    <SILECS-Class name="{project_name}" version="0.1.0" domain="OPERATIONAL" >
        <Acquisition-Block name="MyBlock" generateFesaProperty="true">
            <Acquisition-Register name="myRegister" generateFesaValueItem="true">
                <scalar format="int32"/>
            </Acquisition-Register>
        </Acquisition-Block>
    </SILECS-Class>
</SILECS-Design>"""

    if os.path.isfile(filepath):
        raise Exception(f"Error: There is already a .silecsdesign file available: {os.path.abspath(filepath)}")

    with open(filepath, "w") as f:
        f.write(design_template)

    print(f"\ncreated new silecsdesign: {filepath}")


def create_silecs_deploy(filepath, project_name):
    date = datetime.datetime.today().strftime('%m/%d/%Y')
    deploy_template=f"""<?xml version="1.0" encoding="UTF-8"?>
<SILECS-Deploy silecs-version="{SILECS_VERSION}" created="{date}" updated="{date}" 
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:noNamespaceSchemaLocation="{DEPLOY_SHEMA_PATH}">
    <Information>
        <Owner user-login="{USER}"/>
        <Editor user-login="{USER}"/>
    </Information>
    <Deploy-Unit name="{project_name}" version="0.1.0"/>
    <SilecsDesign silecs-design-name="" silecs-design-version=""/>
        
    <Controller host-name="">
        <Siemens-PLC system="TIA-PORTAL" model="SIMATIC_S7-300" protocol="DEVICE_MODE" base-DB-number="1">
            <Device silecs-device-label="dev0" silecs-design-ref="" fesa-device-name="" fesa-fec-name=""/>
        </Siemens-PLC>
    </Controller>
</SILECS-Deploy>"""

    if os.path.isfile(filepath):
        raise Exception(f"Error: There is already a .silecsdeploy file available: {os.path.abspath(filepath)}")

    with open(filepath, "w") as f:
        f.write(design_template)

    print(f"\ncreated new silecsdeploy: {filepath}")


def create(filepath):
    project_name = get_project_name(filepath)

    if extension == SILECSDESIGN:
        create_silecs_design(filepath, project_name)
    elif extension == SILECSDEPLOY:
        create_silecs_deploy(filepath, project_name)
        

def silecs_create(file_paths):
    for path in file_paths:
        path = os.path.abspath(path)
        try:
            new_file_path = get_silecs_new_file_path(path)
            create(new_file_path)
            print("Created")
        except Exception as e:  
            print(e)

"""
    ####### Generation #######
"""

def silecs_generate_silecsdeploy(silecs_deploy_path):
    project_name = get_project_name(silecs_deploy_path)
    project_dir = get_project_dir(silecs_deploy_path)
    workspace_path = os.path.abspath(os.path.join(project_dir, os.pardir)) 
    fesa_deploy_file = os.path.abspath(os.path.join(project_dir, project_name + ".depoy"))
    SILECS_DEPLOY_VERSION="unused_parameter"

    print(f"generating code for {project_name} ...\n") 
    print(fesa_deploy_file)
    if os.path.isfile(fesa_deploy_file):
        info = f"Please note that old, obsolete xml-elements are not deleted automatically! Sometimes this can lead to an invalid FESA document! \n" \
                f"If this is the case, please remove the obsolete elements by hand. \n" \
                f"Creating a backup of the existing FESA file at: {fesa_deploy_file}.backup \n"
    
        print(info)
        create_backup_file(fesa_deploy_file)
     
    SILECS_CODEGEN_BASE=f"{SILECS_BASE}/silecs-codegen/xml/"
    SILECS_CODEGEN_MIGRATION=f"{SILECS_BASE}/silecs-codegen/xml/migration"
    SILECS_CODEGEN_FESA=f"{SILECS_BASE}/silecs-codegen/xml/fesa/fesa_7_3_0/"

    genparam.genParam(workspace_path, project_name, SILECS_DEPLOY_VERSION, SILECS_VERSION)
    genplcsrc.genPlcSrc(workspace_path, project_name, SILECS_VERSION)
    genduwrapper.genDuWrapper(workspace_path, project_name, SILECS_DEPLOY_VERSION)
    fillFESADeployUnit.fillDeployUnit(workspace_path, project_name, FESA_XSD, FESA_VERSION)

    info = "Please note that you need to add the dependency to the silecs library to your Makefile. \n" \
    	    "Here an example: https://www-acc.gsi.de/wiki/Frontend/Silecs_1_0_0_CodeSnippets#Dependency_in_Makefile.specific !\n"
    print(info)
    print("Code generation finished")


def silecs_generate_silecsdesign(silecs_design_path):

    project_name = get_project_name(silecs_design_path)
    project_dir = get_project_dir(silecs_design_path)
    workspace_path = os.path.abspath(os.path.join(project_dir, os.pardir)) 
    fesa_design_file = os.path.abspath(os.path.join(project_dir, project_name + ".design"))

    print(f"generating code for {project_name} ...\n") 
    print(fesa_design_file)
    if os.path.isfile(fesa_design_file):
        info = f"Please note that old, obsolete xml-elements are not deleted automatically! Sometimes this can lead to an invalid FESA document! \n" \
                f"If this is the case, please remove the obsolete elements by hand. \n" \
                f"Creating a backup of the existing FESA file at: {fesa_design_file}.backup \n"
    
        print(info)
        create_backup_file(fesa_design_file)
     
    info = "Please not that you need to add the dependency to the silecs library to your Makefile. \n" \
    	    "Here an example: https://www-acc.gsi.de/wiki/Frontend/Silecs_1_0_0_CodeSnippets#Dependency_in_Makefile.specific !\n"
    print(info)


    import sys
    sys.path.append(SILECS_CODEGEN_BASE)
    sys.path.append(SILECS_CODEGEN_MIGRATION)
    sys.path.append(SILECS_CODEGEN_FESA)
    
    import generateFesaDesign
    generateFesaDesign.fillDesignFile(FESA_VERSION, project_name, workspace_path, FESA_XSD)
    generateSourceCode.genCppFiles(project_name, workspace_path, silecs_design_path) 

    print("Code generation finished")

def generate(silecs_design_deploy_path):
    extension = get_extension(silecs_design_deploy_path)
    silecs_design_deploy_path = os.path.abspath(silecs_design_deploy_path)

    if extension == SILECSDESIGN:
        silecs_generate_silecsdesign(silecs_design_deploy_path)
    elif extension == SILECSDEPLOY:
        silecs_generate_silecsdeploy(silecs_design_deploy_path)


def silecs_generate(file_paths):
    for path in file_paths:
        try:

            xsd_path = get_schema_path(path)
            val_result = validate(path, xsd_path)
            if val_result:
                print("File is valid")
            else:
                print("File is not valid. Check errors above.")
                continue

            extension = get_extension(path)
            project_name = get_project_name(path)

            generate(path)
        except Exception as e:  
            print(e)

"""
    ####### Parse arguments #######
"""

def _parse_arguments():
    parser = argparse.ArgumentParser(
        description=__description__)

    parser.add_argument(
        "-c",
        "--create",
        nargs="+",
        help="create new .silecsdesign/.silecsdeploy file for the specified FESA .design/.deplox file"
    )

    parser.add_argument(
        "-v",
        "--validate",
        nargs="+",
        help="validate silecs xml file"
    )

    parser.add_argument(
        "-g",
        "--generate",
        nargs="+",
        help="generate source code for silecs xml file"
    )

    parser.add_argument(
        "-d",
        "--diagnostic",
        help="start diagnostic tool for silecs xml file"
    )

    parser.add_argument(
        "-m",
        "--migrate",
        help="migrate silecs design/deploy to new silecs version (TODO)"
    )

    return parser


"""
    ####### Entrz point #######
"""
def run():
    parser = _parse_arguments()

    if len(sys.argv)==1:
        parser.print_help(sys.stderr)

    try:
        options = parser.parse_args()

    except:
        sys.exit(0)

    if options.create:
        silecs_create(options.create)
    if options.validate:
        silecs_validate(options.validate)
    if options.generate:
        silecs_generate(options.generate) 


if __name__ == "__main__":
    run()